# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from datetime import datetime
import os
import copy
import ldap
from lib389._constants import DN_LDBM, DN_CHAIN, DN_PLUGIN, DEFAULT_BENAME, DIRSRV_STATE_ONLINE
from lib389.properties import BACKEND_OBJECTCLASS_VALUE, BACKEND_PROPNAME_TO_ATTRNAME, BACKEND_CHAIN_BIND_DN, \
                              BACKEND_CHAIN_BIND_PW, BACKEND_CHAIN_URLS, BACKEND_PROPNAME_TO_ATTRNAME, BACKEND_NAME, \
                              BACKEND_SUFFIX, BACKEND_SAMPLE_ENTRIES, TASK_WAIT, TASK_WATCH
from lib389.utils import normalizeDN, ensure_str, assert_c, get_default_db_lib
from lib389 import Entry

# Need to fix this ....

from lib389._mapped_object import DSLdapObjects, DSLdapObject, CompositeDSLdapObject
from lib389.mappingTree import MappingTrees
from lib389.exceptions import NoSuchEntryError, InvalidArgumentError
from lib389.replica import Replicas, Changelog
from lib389.cos import (CosTemplates, CosIndirectDefinitions,
                        CosPointerDefinitions, CosClassicDefinitions)

from lib389.index import Index, Indexes, VLVSearches, VLVSearch
from lib389.tasks import ImportTask, ExportTask, Tasks
from lib389.encrypted_attributes import EncryptedAttr, EncryptedAttrs


# This is for sample entry creation.
from lib389.configurations import get_sample_entries

from lib389.lint import DSBLE0001, DSBLE0002, DSBLE0003, DSBLE0004, DSBLE0005, DSBLE0006, DSBLE0007, DSBLE0008, DSVIRTLE0001, DSCLLE0001
from lib389.plugins import USNPlugin
from lib389.dseldif import DSEldif
from lib389._mapped_object_lint import (
    lint_get_attr_val_utf8,
    lint_get_attr_val_utf8_l,
    lint_get_attr_vals_utf8,
    lint_get_attr_vals_utf8_l,
    lint_plugin_enabled,
)


def is_subsuffix_of(sub_parent, be_suffix, all_suffixes):
    """Check if sub_parent indicates this is a sub-suffix of be_suffix.

    Returns True only if be_suffix is the CLOSEST ancestor suffix of sub_parent.
    This prevents a sub-suffix from appearing under multiple ancestors.

    :param sub_parent: The nsslapd-parent-suffix value (lowercase)
    :param be_suffix: The suffix to check against (lowercase)
    :param all_suffixes: Set of all backend suffixes (lowercase)
    :returns: True if be_suffix is the closest ancestor suffix
    """
    if not sub_parent:
        return False
    if sub_parent == be_suffix:
        return True
    if sub_parent in all_suffixes:
        # sub_parent is itself a suffix, will be handled separately
        return False
    if not sub_parent.endswith(',' + be_suffix):
        return False
    # Find the closest (longest) matching suffix for this parent
    best_match = None
    for sfx in all_suffixes:
        if sub_parent == sfx or sub_parent.endswith(',' + sfx):
            if best_match is None or len(sfx) > len(best_match):
                best_match = sfx
    # Only return True if be_suffix is the closest match
    return best_match == be_suffix


class BackendLegacy(object):
    proxied_methods = 'search_s getEntry'.split()

    def __init__(self, conn):
        """@param conn - a DirSrv instance"""
        self.conn = conn
        self.log = conn.log

    def __getattr__(self, name):
        if name in Backend.proxied_methods:
            from lib389 import DirSrv
            return DirSrv.__getattr__(self.conn, name)

    def list(self, suffix=None, backend_dn=None, bename=None):
        """
            Returns a search result of the backend(s) entries with all their
            attributes

            If 'suffix'/'backend_dn'/'benamebase' are specified. It uses
            'backend_dn' first, then 'suffix', then 'benamebase'.

            If neither 'suffix', 'backend_dn' and 'benamebase' are specified,
            it returns all the backend entries

            Get backends by name or suffix

            @param suffix - suffix of the backend
            @param backend_dn - DN of the backend entry
            @param bename - 'commonname'/'cn' of the backend (e.g. 'userRoot')

            @return backend entries

            @raise None
        """

        filt = "(objectclass=%s)" % BACKEND_OBJECTCLASS_VALUE
        if backend_dn:
            self.log.info("List backend %s", backend_dn)
            base = backend_dn
            scope = ldap.SCOPE_BASE
        elif suffix:
            self.log.info("List backend with suffix=%s", suffix)
            base = DN_PLUGIN
            scope = ldap.SCOPE_SUBTREE
            filt = ("(&%s(|(%s=%s)(%s=%s)))" %
                    (filt,
                     BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_SUFFIX], suffix,
                     BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_SUFFIX],
                     normalizeDN(suffix))
                    )
        elif bename:
            self.log.info("List backend 'cn=%s'", bename)
            base = "%s=%s,%s" % (BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_NAME],
                                 bename, DN_LDBM)
            scope = ldap.SCOPE_BASE
        else:
            self.log.info("List all the backends")
            base = DN_PLUGIN
            scope = ldap.SCOPE_SUBTREE

        try:
            ents = self.conn.search_s(base, scope, filt)
        except ldap.NO_SUCH_OBJECT:
            return None

        return ents

    def _readonly(self, bename=None, readonly='on', suffix=None):
        """Put a database in readonly mode
            @param  bename  -   the backend name (eg. addressbook1)
            @param  readonly-   'on' or 'off'

            NOTE: I can ldif2db to a read-only database. After the
                  import, the database will still be in readonly.

            NOTE: When a db is read-only, it seems you need to restart
                  the directory server before creating further
                  agreements or initialize consumers
        """
        if bename and suffix:
            raise ValueError("Specify either bename or suffix")

        if suffix:
            raise NotImplementedError()

        self.conn.modify_s(','.join(('cn=' + bename, DN_LDBM)), [
            (ldap.MOD_REPLACE, BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_READONLY],
             readonly)
        ])

    def delete(self, suffix=None, backend_dn=None, bename=None):
        """
        Deletes the backend entry with the following steps:

        Delete the indexes entries under this backend
        Delete the encrypted attributes entries under this backend
        Delete the encrypted attributes keys entries under this backend

        If a mapping tree entry uses this backend (nsslapd-backend),
        it raise ldap.UNWILLING_TO_PERFORM

        If 'suffix'/'backend_dn'/'benamebase' are specified.
        It uses 'backend_dn' first, then 'suffix', then 'benamebase'.

        @param suffix - suffix of the backend
        @param backend_dn - DN of the backend entry
        @param bename - 'commonname'/'cn' of the backend (e.g. 'userRoot')

        @return None

        @raise ldap.UNWILLING_TO_PERFORM - if several backends match the
                                     argument provided suffix does not
                                     match backend suffix.  It exists a
                                     mapping tree that use that backend


        """

        # First check the backend exists and retrieved its suffix
        be_ents = self.conn.backend.list(suffix=suffix,
                                         backend_dn=backend_dn,
                                         bename=bename)
        if len(be_ents) == 0:
            raise ldap.UNWILLING_TO_PERFORM(
                "Unable to retrieve the backend (%r, %r, %r)" %
                (suffix, backend_dn, bename))
        elif len(be_ents) > 1:
            for ent in be_ents:
                self.log.fatal("Multiple backend match the definition: %s",
                               ent.dn)
            if (not suffix) and (not backend_dn) and (not bename):
                raise ldap.UNWILLING_TO_PERFORM(
                    "suffix and backend DN and backend name are missing")
            raise ldap.UNWILLING_TO_PERFORM(
                "Not able to identify the backend to delete")
        else:
            be_ent = be_ents[0]
            be_suffix = be_ent.getValue(
                BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_SUFFIX])

        # Verify the provided suffix is the one stored in the found backend
        if suffix:
            if normalizeDN(suffix) != normalizeDN(be_suffix):
                raise ldap.UNWILLING_TO_PERFORM(
                    "provided suffix (%s) differs from backend suffix (%s)"
                    % (suffix, be_suffix))

        # now check there is no mapping tree for that suffix
        mt_ents = self.conn.mappingtree.list(suffix=be_suffix)
        if len(mt_ents) > 0:
            raise ldap.UNWILLING_TO_PERFORM(
                "It still exists a mapping tree (%s) for that backend (%s)" %
                (mt_ents[0].dn, be_ent.dn))

        # Now delete the indexes
        found_bename = ensure_str(be_ent.getValue(BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_NAME]))
        if not bename:
            bename = found_bename
        elif bename.lower() != found_bename.lower():
            raise ldap.UNWILLING_TO_PERFORM("Backend name specified (%s) differs from the retrieved one (%s)" % (bename, found_bename))

        self.conn.index.delete_all(bename)

        # finally delete the backend children and the backend itself
        ents = self.conn.search_s(be_ent.dn, ldap.SCOPE_ONELEVEL)
        for ent in ents:
            self.log.debug("Delete entry children %s", ent.dn)
            self.conn.delete_s(ent.dn)

        self.log.debug("Delete backend entry %s", be_ent.dn)
        self.conn.delete_s(be_ent.dn)

        return

    def create(self, suffix=None, properties=None):
        """
            Creates backend entry and returns its dn.

            If the properties 'chain-bind-pwd' and 'chain-bind-dn' and
            'chain-urls' are specified the backend is a chained backend.  A
            chaining backend is created under
                'cn=chaining database,cn=plugins,cn=config'.

            A local backend is created under
                'cn=ldbm database,cn=plugins,cn=config'

            @param suffix - suffix stored in the backend
            @param properties - dictionary with properties values
            supported properties are
                BACKEND_NAME          = 'name'
                BACKEND_READONLY      = 'read-only'
                BACKEND_REQ_INDEX     = 'require-index'
                BACKEND_CACHE_ENTRIES = 'entry-cache-number'
                BACKEND_CACHE_SIZE    = 'entry-cache-size'
                BACKEND_DNCACHE_SIZE  = 'dn-cache-size'
                BACKEND_DIRECTORY     = 'directory'
                BACKEND_DB_DEADLOCK   = 'db-deadlock'
                BACKEND_CHAIN_BIND_DN = 'chain-bind-dn'
                BACKEND_CHAIN_BIND_PW = 'chain-bind-pw'
                BACKEND_CHAIN_URLS    = 'chain-urls'
                BACKEND_SUFFIX        = 'suffix'
                BACKEND_SAMPLE_ENTRIES = 'sample_entries'

            @return backend DN of the created backend

            @raise LDAPError

        """
        def _getBackendName(parent):
            '''
                Use to build a backend name that is not already used
            '''
            index = 1
            while True:
                bename = "local%ddb" % index
                base = ("%s=%s,%s" %
                        (BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_NAME],
                         bename, parent))
                filt = "(objectclass=%s)" % BACKEND_OBJECTCLASS_VALUE
                self.log.debug("_getBackendName: baser=%s : fileter=%s",
                               base, filt)
                try:
                    self.conn.getEntry(base, ldap.SCOPE_BASE, filt)
                except (NoSuchEntryError, ldap.NO_SUCH_OBJECT):
                    self.log.info("backend name will be %s", bename)
                    return bename
                index += 1

        # suffix is mandatory. If may be in the properties
        if isinstance(properties, dict) and properties.get(BACKEND_SUFFIX, None) is not None:
            suffix = properties.get(BACKEND_SUFFIX)
        if not suffix:
            raise ldap.UNWILLING_TO_PERFORM('Missing Suffix')
        else:
            nsuffix = normalizeDN(suffix)

        # Check it does not already exist a backend for that suffix
        if self.conn.verbose:
            self.log.info("Checking suffix %s for existence", suffix)
        ents = self.conn.backend.list(suffix=suffix)
        if len(ents) != 0:
            raise ldap.ALREADY_EXISTS
        # Check if we are creating a local/chained backend
        chained_suffix = (properties and
                          (BACKEND_CHAIN_BIND_DN in properties) and
                          (BACKEND_CHAIN_BIND_PW in properties) and
                          (BACKEND_CHAIN_URLS in properties))
        if chained_suffix:
            self.log.info("Creating a chaining backend")
            dnbase = DN_CHAIN
        else:
            self.log.info("Creating a local backend")
            dnbase = DN_LDBM

        # Get the future backend name
        if properties and BACKEND_NAME in properties:
            cn = properties[BACKEND_NAME]
        else:
            cn = _getBackendName(dnbase)

        # Check the future backend name does not already exists
        # we can imagine having no backends for 'suffix' but having a backend
        # with the same name
        dn = "%s=%s,%s" % (BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_NAME], cn,
                           dnbase)
        ents = self.conn.backend.list(backend_dn=dn)
        if ents:
            raise ldap.ALREADY_EXISTS("Backend already exists with that DN: %s"
                                      % ents[0].dn)

        # All checks are done, Time to create the backend
        try:
            entry = Entry(dn)
            entry.update({
                'objectclass': ['top', 'extensibleObject',
                                BACKEND_OBJECTCLASS_VALUE],
                BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_NAME]: cn,
                BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_SUFFIX]: nsuffix,
            })

            if chained_suffix:
                entry.update(
                    {BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_CHAIN_URLS]:
                     properties[BACKEND_CHAIN_URLS],
                     BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_CHAIN_BIND_DN]:
                     properties[BACKEND_CHAIN_BIND_DN],
                     BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_CHAIN_BIND_PW]:
                     properties[BACKEND_CHAIN_BIND_PW]
                     })

            self.log.debug("adding entry: %r", entry)
            self.conn.add_s(entry)
        except ldap.ALREADY_EXISTS as e:
            self.log.error("Entry already exists: %r", dn)
            raise ldap.ALREADY_EXISTS("%s : %r" % (e, dn))
        except ldap.LDAPError as e:
            self.log.error("Could not add backend entry: %r", dn)
            raise e

        backend_entry = self.conn._test_entry(dn, ldap.SCOPE_BASE)

        return backend_entry

    def getProperties(self, suffix=None, backend_dn=None, bename=None,
                      properties=None):
        raise NotImplementedError

    def setProperties(self, suffix=None, backend_dn=None, bename=None,
                      properties=None):
        raise NotImplementedError

    def toSuffix(self, entry=None, name=None):
        '''
            Return, for a given backend entry, the suffix values.
            Suffix values are identical from a LDAP point of views.
            Suffix values may be surrounded by ", or containing '\'
            escape characters.

            @param entry - LDAP entry of the backend
            @param name  - backend DN

            @result list of values of suffix attribute (aka 'cn')

            @raise ldap.NO_SUCH_OBJECT - in name is invalid DN
                   ValueError - entry does not contains the suffix attribute
                   InvalidArgumentError - if both entry/name are missing
        '''
        attr_suffix = BACKEND_PROPNAME_TO_ATTRNAME[BACKEND_SUFFIX]
        if entry:
            if not entry.hasValue(attr_suffix):
                raise ValueError("Entry has no %s attribute %r" %
                                 (attr_suffix, entry))
            return entry.getValues(attr_suffix)
        elif name:
            filt = "(objectclass=%s)" % BACKEND_OBJECTCLASS_VALUE

            attrs = [attr_suffix]
            ent = self.conn.getEntry(name, ldap.SCOPE_BASE, filt, attrs)
            self.log.debug("toSuffix: %s found by its DN", ent.dn)

            if not ent.hasValue(attr_suffix):
                raise ValueError("Entry has no %s attribute %r" %
                                 (attr_suffix, ent))
            return ent.getValues(attr_suffix)
        else:
            raise InvalidArgumentError("entry or name are mandatory")

    def requireIndex(self, suffix):
        '''
        Should be moved in setProperties
        '''
        entries_backend = self.backend.list(suffix=suffix)
        # assume 1 local backend
        dn = entries_backend[0].dn
        replace = [(ldap.MOD_REPLACE, 'nsslapd-require-index', 'on')]
        self.modify_s(dn, replace)

    @classmethod
    def lint_uid(cls):
        return 'backends'


def _index_entry_attrs_from_dse(instance, bename, attr_name):
    """Load one index entry from ``dse.ldif`` (attribute keys lowercased)."""
    dse = DSEldif(instance)
    dn = "cn={},cn=index,cn={},cn=ldbm database,cn=plugins,cn=config".format(
        attr_name.lower(), bename.lower())
    return dse.get_entry_attrs(dn)


class Backend(DSLdapObject):
    """Backend DSLdapObject with:
    - must attributes = ['cn', 'nsslapd-suffix']
    - RDN attribute is 'cn'

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    _must_attributes = ['nsslapd-suffix', 'cn']

    def __init__(self, instance, dn=None):
        super(Backend, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['nsslapd-suffix', 'cn']
        self._create_objectclasses = ['top', 'extensibleObject', BACKEND_OBJECTCLASS_VALUE]
        self._protected = False
        # Check if a mapping tree for this suffix exists.
        self._mts = MappingTrees(self._instance)

    def lint_uid(self):
        return lint_get_attr_val_utf8_l(self, 'cn')

    def _lint_virt_attrs(self):
        """Check if any virtual attribute are incorrectly indexed"""
        bename = self.lint_uid()
        indexes = self.get_indexes()
        suffix = lint_get_attr_val_utf8(self, 'nsslapd-suffix')
        server_running = self._instance.status()

        # First check nsrole
        try:
            if server_running:
                indexes.get('nsrole')
            else:
                index = DSLdapObject(self._instance,
                                     dn=f'cn=nsrole,cn=index,cn={bename},cn=ldbm database,cn=plugins,cn=config')
                if lint_get_attr_val_utf8_l(index, "objectclass") is None:
                    return

            report = copy.deepcopy(DSVIRTLE0001)
            report['check'] = f'backends:{bename}:virt_attrs'
            report['detail'] = report['detail'].replace('ATTR', 'nsrole')
            report['fix'] = report['fix'].replace('ATTR', 'nsrole')
            report['fix'] = report['fix'].replace('SUFFIX', suffix)
            report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
            report['items'].append(suffix)
            report['items'].append('nsrole')
            yield report
        except:
            pass

        if not server_running:
            # Can not check the database content when the server is not running
            return

        # Check COS next
        for cosDefType in [CosIndirectDefinitions, CosPointerDefinitions, CosClassicDefinitions]:
            defs = cosDefType(self._instance, suffix).list()
            for cosDef in defs:
                attrs = lint_get_attr_val_utf8_l(cosDef, "cosAttribute").split()
                for attr in attrs:
                    if attr in ["default", "override", "operational", "operational-default", "merge-schemes"]:
                        # We are at the end, just break out
                        break
                    try:
                        indexes.get(attr)
                        # If we got here there is an index (bad)
                        report = copy.deepcopy(DSVIRTLE0001)
                        report['check'] = f'backends:{bename}:virt_attrs'
                        report['detail'] = report['detail'].replace('ATTR', attr)
                        report['fix'] = report['fix'].replace('ATTR', attr)
                        report['fix'] = report['fix'].replace('SUFFIX', suffix)
                        report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
                        report['items'].append(suffix)
                        report['items'].append("Class Of Service (COS)")
                        report['items'].append("cosAttribute: " + attr)
                        yield report
                    except:
                        # this is what we hope for
                        pass

    def _lint_search(self):
        """Perform a search and make sure an entry is accessible
        """
        dn = lint_get_attr_val_utf8(self, 'nsslapd-suffix')
        bename = self.lint_uid()
        suffix = DSLdapObject(self._instance, dn=dn)
        try:
            suffix.get_attr_val('objectclass')
        except ldap.NO_SUCH_OBJECT:
            # backend root entry not created yet
            DSBLE0003['items'] = [dn, ]
            DSBLE0003['check'] = f'backends:{bename}:search'
            yield DSBLE0003
        except ldap.LDAPError as e:
            # Some other error
            DSBLE0002['detail'] = DSBLE0002['detail'].replace('ERROR', str(e))
            DSBLE0002['check'] = f'backends:{bename}:search'
            DSBLE0002['items'] = [dn, ]
            yield DSBLE0002

    def _lint_mappingtree(self):
        """Backend lint

        This should check for:
        * missing mapping tree entries for the backend
        * missing indices if we are local and have log access?
        """
        # Check for the missing mapping tree.
        suffix = lint_get_attr_val_utf8_l(self, 'nsslapd-suffix')
        bename = self.lint_uid()
        server_running = self._instance.status()
        mt = None
        try:
            if server_running:
                mt = self._mts.get(suffix)
            else:
                dse = DSEldif(self._instance)
                mapping_trees = dse.get_mapping_trees()
                for mapping_tree in mapping_trees:
                    dn = mapping_tree.replace("dn: ", "")
                    tree_obj = DSLdapObject(self._instance, dn=dn)
                    suffixes = lint_get_attr_vals_utf8_l(tree_obj, 'cn')
                    for suf in suffixes:
                        if suf.lower() == suffix.lower():
                            mt = tree_obj
                            break
            if mt is None or (lint_get_attr_val_utf8_l(mt, 'nsslapd-backend') != bename.lower() and
                              lint_get_attr_val_utf8(mt, 'nsslapd-state') != 'backend'):
                raise ldap.NO_SUCH_OBJECT("We have a matching suffix, but not a backend or correct database name.")
        except ldap.NO_SUCH_OBJECT:
            result = DSBLE0001
            result['check'] = f'backends:{bename}:mappingtree'
            result['items'] = [bename, ]
            yield result


    def _lint_backend_implementation(self):
        backend_impl = self._instance.get_db_lib()
        if backend_impl == 'bdb':
            result = DSBLE0006
            result['items'] = [self.lint_uid()]
            yield result


    def _lint_backend_implementation_cleanup_needed(self):
        db_files = os.listdir(self._instance.ds_paths.db_dir)
        if self._instance.ds_paths.db_home_dir is not None and \
           self._instance.ds_paths.db_home_dir != self._instance.ds_paths.db_dir and \
           os.path.isdir(self._instance.ds_paths.db_home_dir):
            db_files.extend(os.listdir(self._instance.ds_paths.db_home_dir))

        db_files = [f.lower() for f in db_files]
        db_files = sorted(set(db_files))

        mdb_files = ['data.mdb', 'info.mdb', 'lock.mdb']
        bdb_files = ['__db.001', 'dbversion', '__db.003', 'userroot', 'log.0000000001', '__db.002']

        bdb_files_present = all(file in db_files for file in bdb_files)
        mdb_files_present = all(file in db_files for file in mdb_files)

        if bdb_files_present and mdb_files_present:
            result = DSBLE0004
            result['items'] = [self.lint_uid()]
            yield result

    def _lint_backend_implementation_config_attributes(self):
        backend_impl = self._instance.get_db_lib()
        if self._instance.state == DIRSRV_STATE_ONLINE:
            config_attrs = DatabaseConfig(self._instance).get()
        else:
            config_attrs = DatabaseConfig.get_combined_flat_from_dse(self._instance)

        mdb_only_attrs = ['nsslapd-mdb-max-size', 'nsslapd-mdb-max-readers', 'nsslapd-mdb-max-dbs']
        bdb_only_attrs = ['nsslapd-dbcachesize',
                          'nsslapd-dbncache',
                          'nsslapd-db-logdirectory',
                          'nsslapd-db-circular-logging',
                          'nsslapd-db-transaction-wait',
                          'nsslapd-db-checkpoint-interval',
                          'nsslapd-db-compactdb-interval',
                          'nsslapd-db-compactdb-time',
                          'nsslapd-db-transaction-batch-val',
                          'nsslapd-db-transaction-batch-min-wait',
                          'nsslapd-db-transaction-batch-max-wait',
                          'nsslapd-db-logbuf-size',
                          'nsslapd-db-page-size',
                          'nsslapd-db-index-page-size',
                          'nsslapd-db-logfile-size',
                          'nsslapd-db-trickle-percentage',
                          'nsslapd-db-spin-count',
                          'nsslapd-db-verbose',
                          'nsslapd-db-debug',
                          'nsslapd-db-locks',
                          'nsslapd-db-locks-monitoring-enabled',
                          'nsslapd-db-locks-monitoring-threshold',
                          'nsslapd-db-locks-monitoring-pause',
                          'nsslapd-db-named-regions',
                          'nsslapd-db-private-mem',
                          'nsslapd-db-private-import-mem',
                          'nsslapd-db-shm-key',
                          'nsslapd-db-debug-checkpointing',
                          'nsslapd-db-lockdown',
                          'nsslapd-db-tx-max',
                          'nsslapd-online-import-encrypt',
                          'nsslapd-db-deadlock-policy']

        if backend_impl == 'bdb':
            invalid_attrs = [attr for attr in mdb_only_attrs if attr in config_attrs]
        else:  # mdb
            invalid_attrs = [attr for attr in bdb_only_attrs if attr in config_attrs]

        if invalid_attrs:
            result = DSBLE0005
            result['items'] = invalid_attrs
            yield result

    def _lint_cl_trimming(self):
        """Check that cl trimming is at least defined to prevent unbounded growth"""
        bename = self.lint_uid()
        suffix = lint_get_attr_val_utf8(self, 'nsslapd-suffix')
        replicas = Replicas(self._instance)
        try:
            # Check if replication is enabled
            if self._instance.status():
                replicas = replicas.get(suffix)
                cl = Changelog(self._instance, suffix=suffix)
            else:
                # Get dse replicas for this suffix
                dse = DSEldif(self._instance)
                replicas = dse.get_replicas()
                if len(replicas) == 0 or not dse.suffix_replicated(suffix):
                    return
                cl = DSLdapObject(self._instance, dn=f'cn=changelog,cn={bename},cn=ldbm database,cn=plugins,cn=config')
            if lint_get_attr_val_utf8(cl, 'nsslapd-changelogmaxentries') is None and \
               lint_get_attr_val_utf8(cl, 'nsslapd-changelogmaxage') is None:
                report = copy.deepcopy(DSCLLE0001)
                report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
                report['check'] = f'backends:{bename}:cl_trimming'
                yield report
        except:
            # Suffix is not replicated
            self._log.debug(f"_lint_cl_trimming - backend ({suffix}) is not replicated")
            pass

    def _lint_system_indexes(self):
        """Check that system indexes are correctly configured"""
        bename = self.lint_uid()
        suffix = lint_get_attr_val_utf8(self, 'nsslapd-suffix')
        indexes = None
        if self._instance.state == DIRSRV_STATE_ONLINE:
            indexes = self.get_indexes()

        # Default system indexes taken from ldap/servers/slapd/back-ldbm/instance.c
        # Note: entryrdn and ancestorid are internal system indexes that are not
        # exposed in cn=config - they are managed internally by the server.
        # parentid works correctly with both lexicographic and integer ordering,
        # so integerOrderingMatch is not required.
        expected_system_indexes = {
            'parentid': {'types': ['eq'], 'matching_rule': None},
            'objectClass': {'types': ['eq'], 'matching_rule': None},
            'aci': {'types': ['pres'], 'matching_rule': None},
            'nscpEntryDN': {'types': ['eq'], 'matching_rule': None},
            'nsUniqueId': {'types': ['eq'], 'matching_rule': None},
            'nsds5ReplConflict': {'types': ['eq', 'pres'], 'matching_rule': None}
        }

        # Default system indexes taken from ldap/ldif/template-dse.ldif.in
        expected_system_indexes.update({
            'nsCertSubjectDN': {'types': ['eq'], 'matching_rule': None},
            'numsubordinates': {'types': ['pres'], 'matching_rule': None},
            'nsTombstoneCSN': {'types': ['eq'], 'matching_rule': None},
            'targetuniqueid': {'types': ['eq'], 'matching_rule': None}
        })


        # RetroCL plugin creates its own backend with an additonal index for changeNumber
        # See ldap/servers/plugins/retrocl/retrocl_create.c
        if suffix and suffix.lower() == 'cn=changelog':
            expected_system_indexes.update({
                'changeNumber': {'types': ['eq'], 'matching_rule': 'integerOrderingMatch'}
            })

        # USN plugin requires entryusn attribute indexed for equality with integerOrderingMatch rule
        # See ldap/ldif/template-dse.ldif.in
        try:
            usn_plugin = USNPlugin(self._instance)
            if lint_plugin_enabled(usn_plugin):
                expected_system_indexes.update({
                    'entryusn': {'types': ['eq'], 'matching_rule': 'integerOrderingMatch'}
                })
        except Exception as e:
            self._log.debug(f"_lint_system_indexes - Error checking USN plugin: {e}")

        discrepancies = []
        remediation_commands = []
        reindex_attrs = set()

        for attr_name, expected_config in expected_system_indexes.items():
            try:
                index = None
                actual_types = []
                actual_mrs = []
                index_missing = True

                if indexes is not None:
                    try:
                        index = indexes.get(attr_name)
                    except ldap.NO_SUCH_OBJECT:
                        index = None
                    except Exception as e:
                        self._log.debug(f"_lint_system_indexes - Error getting index {attr_name}: {e}")
                        discrepancies.append(f"Unable to check index {attr_name}: {str(e)}")
                        continue
                    index_missing = index is None
                    if index is not None:
                        actual_types = lint_get_attr_vals_utf8(index, 'nsIndexType') or []
                        actual_mrs = lint_get_attr_vals_utf8(index, 'nsMatchingRule') or []
                else:
                    entry = _index_entry_attrs_from_dse(self._instance, bename, attr_name)
                    index_missing = not entry
                    if entry:
                        actual_types = entry.get('nsindextype', []) or []
                        actual_mrs = entry.get('nsmatchingrule', []) or []

                # Check if index exists
                if index_missing:
                    discrepancies.append(f"Missing system index: {attr_name}")
                    # Generate remediation command
                    index_types = ' '.join([f"--index-type {t}" for t in expected_config['types']])
                    cmd = f"dsconf YOUR_INSTANCE backend index add {bename} --attr {attr_name} {index_types}"
                    if expected_config['matching_rule']:
                        cmd += f" --matching-rule {expected_config['matching_rule']}"
                    remediation_commands.append(cmd)
                    reindex_attrs.add(attr_name)  # New index needs reindexing
                else:
                    # Index exists, check configuration
                    # Normalize to lowercase for comparison
                    actual_types = [t.lower() for t in actual_types]
                    expected_types = [t.lower() for t in expected_config['types']]

                    # Check index types
                    missing_types = set(expected_types) - set(actual_types)
                    if missing_types:
                        discrepancies.append(f"Index {attr_name} missing types: {', '.join(missing_types)}")
                        missing_type_args = ' '.join([f"--add-type {t}" for t in missing_types])
                        cmd = f"dsconf YOUR_INSTANCE backend index set {bename} --attr {attr_name} {missing_type_args}"
                        remediation_commands.append(cmd)
                        reindex_attrs.add(attr_name)

                    # Check matching rules
                    expected_mr = expected_config.get('matching_rule')
                    if expected_mr:
                        actual_mrs_lower = [mr.lower() for mr in actual_mrs]
                        if expected_mr.lower() not in actual_mrs_lower:
                            discrepancies.append(f"Index {attr_name} missing matching rule: {expected_mr}")
                            # Add the missing matching rule
                            cmd = f"dsconf YOUR_INSTANCE backend index set {bename} --attr {attr_name} --add-mr {expected_mr}"
                            remediation_commands.append(cmd)
                            reindex_attrs.add(attr_name)

            except Exception as e:
                self._log.debug(f"_lint_system_indexes - Error checking index {attr_name}: {e}")
                discrepancies.append(f"Unable to check index {attr_name}: {str(e)}")

        if discrepancies:
            report = copy.deepcopy(DSBLE0007)
            report['check'] = f'backends:{bename}:system_indexes'
            report['items'] = [suffix]

            expected_indexes_list = []
            for attr_name, config in expected_system_indexes.items():
                types_str = "', '".join(config['types'])
                index_desc = f"- {attr_name}: index type{'s' if len(config['types']) > 1 else ''} '{types_str}'"
                if config['matching_rule']:
                    index_desc += f" with matching rule '{config['matching_rule']}'"
                expected_indexes_list.append(index_desc)

            formatted_expected_indexes = '\n'.join(expected_indexes_list)
            report['detail'] = report['detail'].replace('EXPECTED_INDEXES', formatted_expected_indexes)
            report['detail'] = report['detail'].replace('DISCREPANCIES', '\n'.join([f"- {d}" for d in discrepancies]))

            formatted_commands = '\n'.join([f"    # {cmd}" for cmd in remediation_commands])
            report['fix'] = report['fix'].replace('REMEDIATION_COMMANDS', formatted_commands)

            # Generate specific reindex commands for affected attributes
            if reindex_attrs:
                reindex_commands = []
                for attr in sorted(reindex_attrs):
                    reindex_cmd = f"dsconf YOUR_INSTANCE backend index reindex {bename} --attr {attr}"
                    reindex_commands.append(f"    # {reindex_cmd}")
                formatted_reindex_commands = '\n'.join(reindex_commands)
            else:
                formatted_reindex_commands = "    # No reindexing needed"

            report['fix'] = report['fix'].replace('REINDEX_COMMANDS', formatted_reindex_commands)
            report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
            report['fix'] = report['fix'].replace('BACKEND_NAME', bename)
            yield report

    def _lint_obsolete_entrydn_index(self):
        """Detect obsolete entrydn index left over from pre-entryrdn migrations."""
        bename = self.lint_uid()
        suffix = self.get_attr_val_utf8('nsslapd-suffix')
        issues = []

        try:
            self.get_indexes().get('entrydn')
            issues.append('entrydn index is configured in cn=index')
        except ldap.NO_SUCH_OBJECT:
            pass
        except Exception as e:
            self._log.debug(f"_lint_obsolete_entrydn_index - config check failed: {e}")
            issues.append(f'Unable to check entrydn index configuration: {e}')

        db_dir = os.path.join(self._instance.ds_paths.db_dir, bename)
        if os.path.isdir(db_dir):
            for name in os.listdir(db_dir):
                if name == 'entrydn.db' or name.startswith('entrydn.db'):
                    issues.append(f'entrydn database file present: {name}')

        if not issues:
            return

        report = copy.deepcopy(DSBLE0008)
        report['check'] = f'backends:{bename}:obsolete_entrydn_index'
        report['items'] = [suffix]
        report['detail'] = report['detail'].replace('BACKEND_NAME', bename)
        report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
        report['fix'] = report['fix'].replace('BACKEND_NAME', bename)
        report['fix'] = report['fix'].replace('DB_DIR', db_dir)
        yield report

    def create_sample_entries(self, version):
        """Creates sample entries under nsslapd-suffix value

        :param version: Sample entries version, i.e. 001003006
        :type version: str
        """

        self._log.debug('Requested sample entries at version %s....' % version)
        # Grab the correct sample entry config - remember this is a function ptr.
        centries = get_sample_entries(version)
        # apply it.
        basedn = self.get_attr_val('nsslapd-suffix')
        cent = centries(self._instance, basedn)
        # Now it's built, we can get the version for logging.
        self._log.debug('Creating sample entries at version %s' % cent.version)
        cent.apply()

    def _validate(self, rdn, properties, basedn):
        # We always need to call the super validate first. This way we can
        # guarantee that properties is a dictionary.
        # However, backend can take different properties. One is
        # based on the actual key, value of the object
        # one is the "python magic" types.
        # So we actually have to do the super validation later.
        if properties is None:
            raise ldap.UNWILLING_TO_PERFORM('Invalid request to create. Properties cannot be None')
        if type(properties) != dict:
            raise ldap.UNWILLING_TO_PERFORM("properties must be a dictionary")

        # This is converting the BACKEND_ types to the DS nsslapd- attribute values
        nprops = {}
        for key, value in properties.items():
            try:
                nprops[BACKEND_PROPNAME_TO_ATTRNAME[key]] = [value, ]
            except KeyError:
                # This means, it's not a mapped value, so continue
                nprops[key] = value

        (dn, valid_props) = super(Backend, self)._validate(rdn, nprops, basedn)

        try:
            self._mts.get(ensure_str(valid_props['nsslapd-suffix'][0]))
            raise ldap.UNWILLING_TO_PERFORM("Mapping tree for this suffix exists!")
        except ldap.NO_SUCH_OBJECT:
            pass
        try:
            self._mts.get(ensure_str(valid_props['cn'][0]))
            raise ldap.UNWILLING_TO_PERFORM("Mapping tree for this database exists!")
        except ldap.NO_SUCH_OBJECT:
            pass
        # We have to stash our valid props so that mapping tree can use them ...
        self._nprops_stash = valid_props

        return (dn, valid_props)

    def create(self, dn=None, properties=None, basedn=DN_LDBM, create_mapping_tree=True):
        """Add a new backend entry, create mapping tree,
         and, if requested, sample entries

        :param dn: DN of the new entry
        :type dn: str
        :param properties: Attributes and parameters for the new entry
        :type properties: dict
        :param basedn: Base DN of the new entry
        :type basedn: str
        :param create_mapping_tree: If a related mapping tree node should be created
        :type create_mapping_tree: bool

        :returns: DSLdapObject of the created entry
        """

        sample_entries = False
        parent_suffix = False

        # normalize suffix (remove spaces between comps)
        if dn is not None:
            dn_comps = ldap.dn.explode_dn(dn.lower())
            dn = ",".join(dn_comps)

        if properties is not None:
            dn_comps = ldap.dn.explode_dn(properties['nsslapd-suffix'])
            ndn = ",".join(dn_comps)
            properties['nsslapd-suffix'] = ndn
            sample_entries = properties.pop(BACKEND_SAMPLE_ENTRIES, False)
            parent_suffix = properties.pop('parent', False)

        # Okay, now try to make the backend.
        backend_obj = super(Backend, self).create(dn, properties, basedn)

        # We check if the mapping tree exists in create, so do this *after*
        if create_mapping_tree is True:
            mapping_tree_properties = {
                'cn': self._nprops_stash['nsslapd-suffix'],
                'nsslapd-state': 'backend',
                'nsslapd-backend': self._nprops_stash['cn'],
            }
            if parent_suffix:
                # This is a subsuffix, set the parent suffix
                mapping_tree_properties['nsslapd-parent-suffix'] = parent_suffix

            try:
                self._mts.create(properties=mapping_tree_properties)
            except Exception as e:
                try:
                    backend_obj.delete()
                except Exception as cleanup_error:
                    self._instance.log.error(f"Failed to cleanup backend after mapping tree creation failure: {cleanup_error}")
                raise e

        # We can't create the sample entries unless a mapping tree was installed.
        if sample_entries is not False and create_mapping_tree is True:
            self.create_sample_entries(sample_entries)
        return backend_obj

    def delete(self):
        """Deletes the backend, it's mapping tree and all related indices.
        This can be changed with the self._protected flag!

        :raises: - UnwillingToPerform - if backend is protected
                 - UnwillingToPerform - if nsslapd-state is not 'backend'
        """

        if self._protected:
            raise ldap.UNWILLING_TO_PERFORM("This is a protected backend!")
        # First check if the mapping tree has our suffix still.
        # suffix = self.get_attr_val('nsslapd-suffix')
        bename = self.get_attr_val_utf8('cn')
        try:
            mt = self._mts.get(selector=bename)
            # Assert the type is "backend"
            # Are these the right types....?
            if mt.get_attr_val_utf8_l('nsslapd-state') != 'backend':
                raise ldap.UNWILLING_TO_PERFORM('Can not delete the mapping tree, not for a backend! You may need to delete this backend via cn=config .... ;_; ')

            # Delete replicas first
            try:
                Replicas(self._instance).get(mt.get_attr_val_utf8('cn')).delete()
            except ldap.NO_SUCH_OBJECT:
                # No replica, no problem
                pass

            # Delete our mapping tree if it exists.
            mt.delete()
        except ldap.NO_SUCH_OBJECT:
            # Righto, it's already gone! Do nothing ...
            pass

        # Now remove our children, this is all ldbm config
        self._instance.delete_branch_s(self._dn, ldap.SCOPE_SUBTREE)

    def get_suffix(self):
        return self.get_attr_val_utf8_l('nsslapd-suffix')

    def disable(self):
        # Disable backend (mapping tree)
        suffix = self.get_attr_val_utf8_l('nsslapd-suffix')
        mt = self._mts.get(suffix)
        mt.set('nsslapd-nsstate', 'Disabled')

    def enable(self):
        # Enable Backend (mapping tree)
        suffix = self.get_attr_val_utf8_l('nsslapd-suffix')
        mt = self._mts.get(suffix)
        mt.set('nsslapd-nsstate', 'Backend')

    def get_mapping_tree(self):
        suffix = self.get_attr_val_utf8('nsslapd-suffix')
        return self._mts.get(suffix)

    def get_monitor(self):
        """Get a MonitorBackend(DSLdapObject) for the backend"""
        # We need to be a factor to the backend monitor
        from lib389.monitor import MonitorBackend

        monitor = MonitorBackend(instance=self._instance, dn="cn=monitor,%s" % self._dn)
        return monitor

    def get_indexes(self):
        """Get an Indexes(DSLdapObject) for the backend"""

        indexes = Indexes(self._instance, basedn="cn=index,%s" % self._dn)
        return indexes

    def get_index(self, attr_name):
        for index in self.get_indexes().list():
            idx_name = index.get_attr_val_utf8_l('cn')
            if idx_name == attr_name.lower():
                return index
        return None

    def del_index(self, attr_name):
        for index in self.get_indexes().list():
            idx_name = index.get_attr_val_utf8_l('cn')
            if idx_name == attr_name.lower():
                index.delete()
                return
        raise ValueError("Can not delete index because it does not exist")

    def add_index(self, attr_name, types, matching_rules=None, reindex=False):
        """ Add an index.

        :param attr_name - name of the attribute to index
        :param types - a List of index types(eq, pres, sub, approx)
        :param matching_rules - a List of matching rules for the index
        :param reindex - If set to True then index the attribute after creating it.
        """

        # Reject adding an index for a virtual attribute
        virt_attr_list = ['nsrole']
        for cosDefType in [CosIndirectDefinitions, CosPointerDefinitions, CosClassicDefinitions]:
            defs = cosDefType(self._instance, self.get_suffix()).list()
            for cosDef in defs:
                attrs = lint_get_attr_val_utf8_l(cosDef, "cosAttribute").split()
                for attr in attrs:
                    if attr in ["default", "override", "operational", "operational-default", "merge-schemes"]:
                        # We are at the end, just break out
                        break
                    virt_attr_list.append(attr)
        if attr_name.lower() in virt_attr_list:
            raise ValueError(f"You should not index a virtual attribute ({attr_name})")

        new_index = Index(self._instance)
        props = {'cn': attr_name,
                 'nsSystemIndex': 'False',
                 'nsIndexType': types,
                 }
        if matching_rules is not None:
            mrs = []
            for mr in matching_rules:
                mrs.append(mr)
            # Only add if there are actually rules present in the list.
            if len(mrs) > 0:
                props['nsMatchingRule'] = mrs
        new_index.create(properties=props, basedn="cn=index," + self._dn)

        if reindex:
            self.reindex(attr_name)

    def reindex(self, attrs=None, wait=False, watch=False):
        """Reindex the attributes for this backend
        :param attrs - an optional list of attributes to index
        :param wait - Set to true to wait for task to complete
        """
        args = {TASK_WAIT: wait, TASK_WATCH: watch}

        bename = self.get_attr_val_utf8('cn')
        reindex_task = Tasks(self._instance)
        reindex_task.reindex(benamebase=bename, attrname=attrs, args=args)

    def get_encrypted_attrs(self, just_names=False):
        """Get a list of the excrypted attributes
        :param just_names - If True only the encrypted attribute names are returned (instead of the full attribute entry)
        :returns - a list of attributes
        """
        attrs = EncryptedAttrs(self._instance, basedn=self._dn).list()
        if just_names:
            results = []
            for attr in attrs:
                results.append(attr.get_attr_val_utf8_l('cn'))
            return results
        else:
            return attrs

    def add_encrypted_attr(self, attr_name):
        """Add an encrypted attribute
        :param attr_name - name of the new encrypted attribute
        """
        new_attr = EncryptedAttr(self._instance)
        new_attr.create(basedn="cn=encrypted attributes," + self._dn, properties={'cn': attr_name,'nsEncryptionAlgorithm': 'AES'})

    def del_encrypted_attr(self, attr_name):
        """Delete encrypted attribute
        :param attr_name - Name of the encrypted attribute to delete
        """
        enc_attrs = EncryptedAttrs(self._instance, basedn="cn=encrypted attributes," + self._dn).list()
        for enc_attr in enc_attrs:
            attr = enc_attr.get_attr_val_utf8_l('cn').lower()
            if attr_name == attr.lower():
                enc_attr.delete()
                break

    def import_ldif(self, ldifs, chunk_size=None, encrypted=False, gen_uniq_id=None, only_core=False,
                    include_suffixes=None, exclude_suffixes=None):
        """Do an import of the suffix"""

        bs = Backends(self._instance)
        task = bs.import_ldif(self.rdn, ldifs, chunk_size, encrypted, gen_uniq_id, only_core,
                              include_suffixes, exclude_suffixes)
        return task

    def export_ldif(self, ldif=None, use_id2entry=False, encrypted=False, min_base64=False, no_uniq_id=False,
                    replication=False, not_folded=False, no_seq_num=False, include_suffixes=None, exclude_suffixes=None):
        """Do an export of the suffix"""

        bs = Backends(self._instance)
        task = bs.export_ldif(self.rdn, ldif, use_id2entry, encrypted, min_base64, no_uniq_id,
                              replication, not_folded, no_seq_num, include_suffixes, exclude_suffixes)
        return task

    def get_vlv_searches(self, vlv_name=None):
        """Return the VLV seaches for this backend, or return a specific search
        :param vlv_name - name of a VLV search entry to return
        :returns - A list of VLV searches or a single VLV sarch entry
        """
        vlv_searches = VLVSearches(self._instance, basedn=self._dn).list()
        if vlv_name is None:
            return vlv_searches

        # return specific search
        for vlv in vlv_searches:
            search_name = vlv.get_attr_val_utf8_l('cn')
            if search_name == vlv_name.lower():
                return vlv

        # No match
        raise ValueError("Failed to find VLV search entry")

    def add_vlv_search(self, vlvname, props, reindex=False):
        """Add a VLV search entry
        :param: vlvname - Name of the new VLV search entry
        :props - A dict of the attribute value pairs for the VLV search entry
        :param - reindex - Set to True to index the new attribute right away
        """
        basedn = self._dn
        vlv = VLVSearch(instance=self._instance)
        vlv.create(rdn="cn=" + vlvname, properties=props, basedn=basedn)

    def get_sub_suffixes(self):
        """Return a list of Backend's that are sub-suffixes of this backend.
        :returns: A list of Backend instances that are sub-suffixes
        """
        subsuffixes = []
        top_be_suffix = self.get_attr_val_utf8_l('nsslapd-suffix')
        if not top_be_suffix:
            return subsuffixes

        mts = self._mts.list()
        be_insts = Backends(self._instance).list()
        all_suffixes = {be.get_attr_val_utf8_l('nsslapd-suffix') for be in be_insts}

        for mt in mts:
            parent_suffix = mt.get_attr_val_utf8_l('nsslapd-parent-suffix')
            if parent_suffix is None:
                continue

            if is_subsuffix_of(parent_suffix, top_be_suffix, all_suffixes):
                child_suffix = mt.get_attr_val_utf8_l('cn')
                for be in be_insts:
                    if child_suffix == be.get_attr_val_utf8_l('nsslapd-suffix'):
                        subsuffixes.append(be)
                        break
        return subsuffixes

    def get_cos_indirect_defs(self):
        return CosIndirectDefinitions(self._instance, self._dn).list()

    def get_cos_pointer_defs(self):
        return CosPointerDefinitions(self._instance, self._dn).list()

    def get_cos_classic_defs(self):
        return CosClassicDefinitions(self._instance, self._dn).list()

    def get_cos_templates(self):
        return CosTemplates(self._instance, self._dn).list()

    def get_state(self):
        suffix = self.get_attr_val_utf8('nsslapd-suffix')
        try:
            mt = self._mts.get(suffix)
        except ldap.NO_SUCH_OBJECT:
            raise ValueError("Backend missing mapping tree entry, unable to get state")
        return mt.get_attr_val_utf8('nsslapd-state')

    def set_state(self, new_state):
        new_state = new_state.lower()
        suffix = self.get_attr_val_utf8('nsslapd-suffix')
        try:
            mt = self._mts.get(suffix)
        except ldap.NO_SUCH_OBJECT:
            raise ValueError("Backend missing mapping tree entry, unable to set configuration")

        if new_state not in ['backend', 'disabled',  'referral',  'referral on update']:
            raise ValueError(f"Invalid backend state {new_state}, value must be one of the following: 'backend', 'disabled',  'referral',  'referral on update'")

        # Can not change state of replicated backend
        replicas = Replicas(self._instance)
        try:
            # Check if replication is enabled
            replicas.get(suffix)
            raise ValueError("Can not change the backend state of a replicated suffix")
        except ldap.NO_SUCH_OBJECT:
            pass

        # Ok, change the state
        mt.replace('nsslapd-state', new_state)


class Backends(DSLdapObjects):
    """DSLdapObjects that represents DN_LDBM base DN
    This only does ldbm backends. Chaining backends are a special case
    of this, so they can be subclassed off.

    :param instance: An instance
    :type instance: lib389.DirSrv
    """

    def __init__(self, instance, basedn=None):
        # Basedn has to be here, despite not being used to satisfy
        # cli_base _generic_create.
        super(Backends, self).__init__(instance=instance)
        self._objectclasses = [BACKEND_OBJECTCLASS_VALUE]
        self._filterattrs = ['cn', 'nsslapd-suffix', 'nsslapd-directory']
        self._childobject = Backend
        self._basedn = DN_LDBM

    def list(self, paged_search=None, paged_critical=True, full_dn=False):
        """List backends via LDAP when online; read ``dse.ldif`` when offline (e.g. healthcheck)."""
        if self._instance.state != DIRSRV_STATE_ONLINE:
            return self._list_from_dse(paged_search=paged_search, paged_critical=paged_critical,
                                       full_dn=full_dn)
        return super(Backends, self).list(paged_search=paged_search, paged_critical=paged_critical,
                                          full_dn=full_dn)

    def _list_from_dse(self, paged_search=None, paged_critical=True, full_dn=False):
        dse = DSEldif(self._instance)
        names = dse.get_backends()
        insts = []
        for name in names:
            dn = f"cn={name},cn=ldbm database,cn=plugins,cn=config"
            if full_dn:
                insts.append(dn)
            else:
                insts.append(Backend(self._instance, dn=dn))
        return insts

    @classmethod
    def lint_uid(cls):
        return 'backends'

    def import_ldif(self, be_name, ldifs, chunk_size=None, encrypted=False, gen_uniq_id=None, only_core=False,
                    include_suffixes=None, exclude_suffixes=None):
        """Do an import of the suffix"""

        if not ldifs:
            raise ValueError("import_ldif: LDIF filename is missing")
        ldif_paths = []
        for ldif in list(ldifs):
            if not ldif.startswith("/"):
                if ldif.endswith(".ldif"):
                    ldif = os.path.join(self._instance.ds_paths.ldif_dir, ldif)
                else:
                    ldif = os.path.join(self._instance.ds_paths.ldif_dir, "%s.ldif" % ldif)
            ldif_paths.append(ldif)

        task = ImportTask(self._instance)
        task_properties = {'nsInstance': be_name,
                           'nsFilename': ldif_paths}
        if include_suffixes is not None:
            task_properties['nsIncludeSuffix'] = include_suffixes
        if exclude_suffixes is not None:
            task_properties['nsExcludeSuffix'] = exclude_suffixes
        if encrypted:
            task_properties['nsExportDecrypt'] = 'true'
        if only_core:
            task_properties['nsImportIndexAttrs'] = 'false'
        if chunk_size is not None:
            task_properties['nsImportChunkSize'] = chunk_size
        if gen_uniq_id is not None:
            if gen_uniq_id in ("none", "empty") or gen_uniq_id.startswith("deterministic"):
                raise ValueError("'gen_uniq_id should be none (no unique ID) |"
                                 "empty (time-based ID) | deterministic namespace (name-based ID)")
            task_properties['nsUniqueIdGenerator'] = gen_uniq_id

        task.create(properties=task_properties)

        return task

    def export_ldif(self, be_names, ldif=None, use_id2entry=False, encrypted=False, min_base64=False, no_dump_uniq_id=False,
                    replication=False, not_folded=False, no_seq_num=False, include_suffixes=None, exclude_suffixes=None):
        """Do an export of the suffix"""

        task = ExportTask(self._instance)
        task_properties = {'nsInstance': be_names}
        if ldif == "":
            ldif = None
        if ldif is not None and not ldif.startswith("/"):
            if ldif.endswith(".ldif"):
                task_properties['nsFilename'] = os.path.join(self._instance.ds_paths.ldif_dir, ldif)
            else:
                task_properties['nsFilename'] = os.path.join(self._instance.ds_paths.ldif_dir, "%s.ldif" % ldif)
        elif ldif is not None and ldif.startswith("/"):
            if ldif.endswith(".ldif"):
                task_properties['nsFilename'] = ldif
            else:
                task_properties['nsFilename'] = "%s.ldif" % ldif
        else:
            tnow = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
            task_properties['nsFilename'] = os.path.join(self._instance.ds_paths.ldif_dir,
                                                         "%s-%s-%s.ldif" % (self._instance.serverid,
                                                                            "-".join(be_names), tnow))
        if include_suffixes is not None:
            task_properties['nsIncludeSuffix'] = include_suffixes
        if exclude_suffixes is not None:
            task_properties['nsExcludeSuffix'] = exclude_suffixes
        if use_id2entry:
            task_properties['nsUseId2Entry'] = 'true'
        if encrypted:
            task_properties['nsExportDecrypt'] = 'true'
        if replication:
            task_properties['nsExportReplica'] = 'true'
        if min_base64:
            task_properties['nsMinimalEncoding'] = 'true'
        if not_folded:
            task_properties['nsNoWrap'] = 'true'
        if no_dump_uniq_id:
            task_properties['nsDumpUniqId'] = 'false'
        if no_seq_num:
            task_properties['nsPrintKey'] = 'false'

        task = task.create(properties=task_properties)
        return task

    def delete_all_dangerous(self):
        """
        Delete all backends. This deletes from longest to shortest suffix
        to ensure correct delete ordering.
        """
        for be in sorted(self.list(), key=lambda be: len(be.get_suffix()), reverse=True):
            be.delete()

    def get_backend(self, suffix):
        """
        Return the backend associated with the provided suffix.
        """
        suffix_l = suffix.lower()
        for be in self.list():
            if be.get_attr_val_utf8_l('nsslapd-suffix') == suffix_l:
                return be
        return None


class DatabaseConfig(DSLdapObject):
    """Backend Database configuration

    The entire database configuration consists of the  main global configuration entry,
    and the underlying DB library configuration: whither BDB or LMDB.  The combined
    configuration should be presented as a single entity so the end user does not need
    to worry about what library is being used, and just focus on the configuration.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    _GLOBAL_ATTRS = [
        'nsslapd-lookthroughlimit',
        'nsslapd-mode',
        'nsslapd-idlistscanlimit',
        'nsslapd-directory',
        'nsslapd-import-cachesize',
        'nsslapd-idl-switch',
        'nsslapd-search-bypass-filter-test',
        'nsslapd-search-use-vlv-index',
        'nsslapd-exclude-from-export',
        'nsslapd-serial-lock',
        'nsslapd-pagedlookthroughlimit',
        'nsslapd-pagedidlistscanlimit',
        'nsslapd-rangelookthroughlimit',
        'nsslapd-backend-opt-level',
        'nsslapd-backend-implement',
        'nsslapd-db-durable-transaction',
        'nsslapd-search-bypass-filter-test',
        'nsslapd-serial-lock',
        'nsslapd-dynamic-lists-enabled',
        'nsslapd-dynamic-lists-attr',
        'nsslapd-dynamic-lists-oc',
        'nsslapd-dynamic-lists-url-attr',
    ]
    _DB_ATTRS = {
        'bdb':
            [
                'nsslapd-dbcachesize',
                'nsslapd-db-logdirectory',
                'nsslapd-db-home-directory',
                'nsslapd-db-transaction-wait',
                'nsslapd-db-checkpoint-interval',
                'nsslapd-db-compactdb-interval',
                'nsslapd-db-compactdb-time',
                'nsslapd-db-page-size',
                'nsslapd-db-transaction-batch-val',
                'nsslapd-db-transaction-batch-min-wait',
                'nsslapd-db-transaction-batch-max-wait',
                'nsslapd-db-logbuf-size',
                'nsslapd-db-locks',
                'nsslapd-db-locks-monitoring-enabled',
                'nsslapd-db-locks-monitoring-threshold',
                'nsslapd-db-locks-monitoring-pause',
                'nsslapd-db-private-import-mem',
                'nsslapd-import-cache-autosize',
                'nsslapd-cache-autosize',
                'nsslapd-cache-autosize-split',
                'nsslapd-import-cachesize',
                'nsslapd-db-deadlock-policy',
            ],
        'mdb': [
                'nsslapd-mdb-max-size',
                'nsslapd-mdb-max-readers',
                'nsslapd-mdb-max-dbs',
                'nsslapd-cache-autosize',
            ]
    }

    def __init__(self, instance, dn="cn=config,cn=ldbm database,cn=plugins,cn=config"):
        super(DatabaseConfig, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['cn']
        self._global_attrs = DatabaseConfig._GLOBAL_ATTRS
        self._db_attrs = DatabaseConfig._DB_ATTRS
        self._create_objectclasses = ['top', 'extensibleObject']
        self._protected = True
        # This could be "bdb" or "mdb", use what we have configured in the global config
        self._db_lib = self.get_attr_val_utf8_l('nsslapd-backend-implement')
        self._dn = "cn=config,cn=ldbm database,cn=plugins,cn=config"
        self._db_dn = f"cn={self._db_lib},cn=config,cn=ldbm database,cn=plugins,cn=config"
        self._globalObj = DSLdapObject(self._instance, dn=self._dn)
        self._dbObj = DSLdapObject(self._instance, dn=self._db_dn)
        # Assert there is no overlap in different config sets
        assert_c(len(set(self._global_attrs).intersection(set(self._db_attrs['bdb']), set(self._db_attrs['mdb']))) == 0)

    @classmethod
    def get_combined_flat_from_dse(cls, instance):
        """Return the same merged attribute dict as :meth:`get`, from ``dse.ldif`` (no LDAP)."""
        dse = DSEldif(instance)
        dn_global = "cn=config,cn=ldbm database,cn=plugins,cn=config"
        ag = dse.get_entry_attrs(dn_global.lower()) or {}
        db_impl = (ag.get('nsslapd-backend-implement') or [None])[0]
        if db_impl is not None:
            db_impl = db_impl.strip().lower()
        if not db_impl:
            db_impl = get_default_db_lib()
        if db_impl not in cls._DB_ATTRS:
            db_impl = get_default_db_lib()
        dn_db = f"cn={db_impl},cn=config,cn=ldbm database,cn=plugins,cn=config"
        attrs_db = dse.get_entry_attrs(dn_db.lower()) or {}
        combined = {}
        for k in cls._GLOBAL_ATTRS:
            combined[k] = ag.get(k, [])
        for k in cls._DB_ATTRS[db_impl]:
            combined[k] = attrs_db.get(k, [])
        return combined

    def get(self):
        """Get the combined config entries"""
        # Get and combine both sets of attributes
        global_attrs = self._globalObj.get_attrs_vals_utf8(self._global_attrs)
        db_attrs = self._dbObj.get_attrs_vals_utf8(self._db_attrs[self._db_lib])
        combined_attrs = {**global_attrs, **db_attrs}
        return combined_attrs

    def display(self):
        """Display the combined configuration"""
        global_attrs = self._globalObj.get_attrs_vals_utf8(self._global_attrs)
        db_attrs = self._dbObj.get_attrs_vals_utf8(self._db_attrs[self._db_lib])
        combined_attrs = {**global_attrs, **db_attrs}
        for (k, vo) in combined_attrs.items():
            if len(vo) == 0:
                vo = ""
            else:
                vo = vo[0]
            self._instance.log.info(f'{k}: {vo}')

    def get_db_lib(self):
        """Return the backend library, bdb, mdb, etc"""
        return self._db_lib

    def set(self, value_pairs):
        for attr, val in value_pairs:
            attr = attr.lower()
            if attr in self._global_attrs:
                global_config = DSLdapObject(self._instance, dn=self._dn)
                global_config.replace(attr, val)
            elif attr in self._db_attrs['bdb']:
                db_config = DSLdapObject(self._instance, dn=self._db_dn)
                db_config.replace(attr, val)
            elif attr in self._db_attrs['mdb']:
                db_config = DSLdapObject(self._instance, dn=self._db_dn)
                db_config.replace(attr, val)
            else:
                # Unknown attribute
                raise ValueError("Can not update database configuration with unknown attribute: " + attr)


class BackendSuffixView(CompositeDSLdapObject):
    """ Composite view between backend and mapping tree entries
        used by: dsconf instance backend suffix ...
    """

    def __init__(self, instance, be):
        super(BackendSuffixView, self).__init__(instance, be.dn)
        be_args = [
            'nsslapd-cachememsize',
            'nsslapd-cachesize',
            'nsslapd-cache-preserved-entries',
            'nsslapd-dncachememsize',
            'nsslapd-readonly',
            'nsslapd-require-index',
            'nsslapd-suffix'
        ]
        mt_args = [
            'orphan',
            'nsslapd-state',
            'nsslapd-referral',
        ]
        mt = be._mts.get(be.get_suffix())
        self.add_component(be, be_args)
        self.add_component(mt, mt_args)

    def get_state(self):
        return self.get_attr_val_utf8('nsslapd-state')

    def set_state(self, new_state):
        new_state = new_state.lower()
        suffix = self.get_attr_val_utf8('nsslapd-suffix')

        if new_state not in ['backend', 'disabled',  'referral',  'referral on update']:
            raise ValueError(f"Invalid backend state {new_state}, value must be one of the following: 'backend', 'disabled',  'referral',  'referral on update'")

        # Can not change state of replicated backend
        replicas = Replicas(self._instance)
        try:
            # Check if replication is enabled
            replicas.get(suffix)
            raise ValueError("Can not change the backend state of a replicated suffix")
        except ldap.NO_SUCH_OBJECT:
            pass

        # Ok, change the state
        self.set('nsslapd-state', new_state)



#!/usr/bin/python3

# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

##################################################
#
# Usage: when using gdb, these commands are automatically loaded with ns-slapd.
#
# else, from inside gdb run "source /path/to/ns-slapd-gdb.py"
#

import itertools
import re
import gdb
from gdb.FrameDecorator import FrameDecorator


class DSAccessLog (gdb.Command):
    """Display the Directory Server access log."""
    def __init__ (self):
        super (DSAccessLog, self).__init__ ("ds-access-log", gdb.COMMAND_DATA)

    def invoke (self, arg, from_tty):
        print('===== BEGIN ACCESS LOG =====')
        gdb.execute('set print elements 0')
        o = gdb.execute('p loginfo.log_access_buffer.top', to_string=True)
        for l in o.split('\\n'):
            print(l)
        print('===== END ACCESS LOG =====')


class DSBacktrace(gdb.Command):
    """Display a filtered backtrace"""
    def __init__ (self):
        super (DSBacktrace, self).__init__ ("ds-backtrace", gdb.COMMAND_DATA)

    def _parse_thread_state(self, lwpid, tid):
        # Stash the BT output
        o = gdb.execute('bt', to_string=True)
        # Get to the top of the frame stack.
        gdb.newest_frame()
        # Now work our way down.
        backtrace = []
        cur_frame = gdb.selected_frame()
        while cur_frame is not None:
            backtrace.append(cur_frame.name())
            cur_frame = cur_frame.older()
        # Dicts can't use lists as keys, so we need to squash this to a string.

        formatted = ['???' if x is None else str(x) for x in backtrace]
        s_backtrace = ' '.join(formatted)
        # Have we seen this trace before?
        if s_backtrace not in self._stack_maps:
            # Make it!
            self._stack_maps[s_backtrace] = []
        # Add it to the set.
        self._stack_maps[s_backtrace].append({'gtid': tid, 'lwpid': lwpid, 'bt': o}  )

    def invoke(self, arg, from_tty):
        print('===== BEGIN ACTIVE THREADS =====')

        # Reset our thread maps.
        self._stack_maps = {}

        inferiors = gdb.inferiors()
        for inferior in inferiors:
            threads = inferior.threads()
            for thread in threads:
                (tpid, lwpid, tid) = thread.ptid
                gtid = thread.num
                thread.switch()
                self._parse_thread_state(lwpid, gtid)

        for m in self._stack_maps:
            # Take a copy of the bt
            o = self._stack_maps[m][0]['bt']
            # Print every thread and id.
            for t in self._stack_maps[m]:
                print("Thread %s (LWP %s))" % (t['gtid'], t['lwpid']))
            # Print the trace
            print(o)

        print('===== END ACTIVE THREADS =====')


class DSIdleFilterDecorator(FrameDecorator):
    def __init__(self, fobj):
        super(DSIdleFilterDecorator, self).__init__(fobj)

    def function(self):
        frame = self.inferior_frame()
        name = str(frame.name())
        if frame.name() is None:
            name = '???'

        if name == 'connection_wait_for_new_work' or name == 'work_q_wait':
            name = "[IDLE THREAD] " + name

        return name


class DSIdleFilter():
    def __init__(self):
        self.name = "DSIdleFilter"
        self.priority = 100
        self.enabled = True
        # Register this frame filter with the global frame_filters
        # dictionary.
        gdb.frame_filters[self.name] = self

    def filter(self, frame_iter):
        # Just return the iterator.
        if hasattr(itertools, 'imap'):
            frame_iter = itertools.imap(DSIdleFilterDecorator, frame_iter)
        else:
            frame_iter = map(DSIdleFilterDecorator, frame_iter)
        return frame_iter


class DSEntryPrint (gdb.Command):
    """Display a Slapi_Entry"""
    def __init__(self):
        super (DSEntryPrint, self).__init__("ds-entry-print", gdb.COMMAND_DATA)

    def _display_values(self, a_name, va_ptr, num):
        inum = int(num)

        for i in range(0, inum):
            value = va_ptr[i].dereference()
            bv = value['bv']['bv_val']
            if bv == 0:
                print("%s: X 0" % a_name)
            else:
                print("%s: %s" % (a_name, bv.string()))


    def _display_attrs(self, start):
        if start.address == 0:
            return
        attr = start.dereference()
        while True:
            name = attr['a_type'].string()
            # print(dir(name))
            va = attr['a_present_values']['va']
            num = attr['a_present_values']['num']
            self._display_values(name, va, num)
            # Now loop
            n = attr['a_next']
            if n == 0:
                return
            attr = n.dereference()


    def invoke (self, arg, from_tty):
        gdb.newest_frame()
        cur_frame = gdb.selected_frame()
        entry_val = cur_frame.read_var(arg)
        entry_root = entry_val.dereference()
        entry_sdn = entry_root['e_sdn']['ndn']
        # Display the SDN
        print("Display Slapi_Entry: %s" % entry_sdn.string())
        # Display the attributes.
        entry_attrs = entry_root['e_attrs']
        self._display_attrs(entry_attrs)

DSAccessLog()
DSBacktrace()
DSIdleFilter()
DSEntryPrint()





# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
from lib389._constants import *
from lib389 import DirSrv, Entry
import pytest
from lib389.utils import ensure_bytes, ensure_str

import sys
MAJOR, MINOR, _, _, _ = sys.version_info

INSTANCE_PORT = 54321
INSTANCE_SERVERID = 'standalone'
# INSTANCE_PREFIX   = None


class TopologyStandalone(object):
    def __init__(self, standalone):
        standalone.open()
        self.standalone = standalone


@pytest.fixture(scope="module")
def topology(request):
    standalone = DirSrv(verbose=False)
    standalone.log.debug("Instance allocated")
    args = {SER_HOST: LOCALHOST,
            SER_PORT: INSTANCE_PORT,
            # SER_DEPLOYED_DIR:  INSTANCE_PREFIX,
            SER_SERVERID_PROP: INSTANCE_SERVERID}
    standalone.allocate(args)
    if standalone.exists():
        standalone.delete()
    standalone.create()
    standalone.open()

    def fin():
        standalone.delete()
    request.addfinalizer(fin)

    return TopologyStandalone(standalone)


@pytest.fixture(scope="module")
def complex_aci(topology):
    ACI_TARGET = ('(targetfilter ="(ou=groups)")(targetattr ="uniqueMember '
                  '|| member")')
    ACI_ALLOW = ('(version 3.0; acl "Allow test aci";allow (read, search, '
                 'write)')
    ACI_SUBJECT = ('(userdn="ldap:///dc=example,dc=com??sub?(ou=engineering)" '
                   'and userdn="ldap:///dc=example,dc=com??sub?(manager=uid='
                   'wbrown,ou=managers,dc=example,dc=com) || ldap:///dc=examp'
                   'le,dc=com??sub?(manager=uid=tbrown,ou=managers,dc=exampl'
                   'e,dc=com)" );)')
    ACI_BODY = ACI_TARGET + ACI_ALLOW + ACI_SUBJECT

    gentry = Entry('cn=testgroup,%s' % DEFAULT_SUFFIX)
    gentry.setValues('objectclass', 'top', 'extensibleobject')
    gentry.setValues('cn', 'testgroup')
    gentry.setValues('aci', ACI_BODY)

    topology.standalone.add_s(gentry)

    return ensure_str(ACI_BODY)


def test_aci(topology, complex_aci):
    """Checks content of previously added aci"""

    acis = topology.standalone.aci.list('cn=testgroup,%s' % DEFAULT_SUFFIX)
    assert len(acis) == 1
    aci = acis[0]
    print(aci.acidata)
    assert aci.acidata == {
        'allow': [{'values': ['read', 'search', 'write']}],
        'target': [], 'targetattr': [{'values': ['uniqueMember', 'member'],
                                      'equal': True}],
        'targattrfilters': [],
        'deny': [],
        'acl': [{'values': ['Allow test aci']}],
        'deny_raw_bindrules': [],
        'targetattrfilters': [],
        'allow_raw_bindrules': [{'values': [(
            'userdn="ldap:///dc=example,dc=com??sub?(ou=engineering)" and'
            ' userdn="ldap:///dc=example,dc=com??sub?(manager=uid=wbrown,'
            'ou=managers,dc=example,dc=com) || ldap:///dc=example,dc=com'
            '??sub?(manager=uid=tbrown,ou=managers,dc=example,dc=com)" ')]}],
        'targetfilter': [{'values': ['(ou=groups)'], 'equal': True}],
        'targetscope': [],
        'version 3.0;': [],
        'rawaci': complex_aci
    }
    assert aci.getRawAci() == complex_aci


def test_aci_lint(topology, complex_aci):
    """Checks content of previously added aci"""

    # Make sure we actually have acis to check!
    acis = topology.standalone.aci.list('cn=testgroup,%s' % DEFAULT_SUFFIX)
    # print(acis)
    assert len(acis) == 1

    # Actually test them
    (result, detail) = topology.standalone.aci.lint(DEFAULT_SUFFIX)
    # What is the point of testing perfect acis! We need this to fail!
    assert (result is False)
    # print(topology.standalone.aci.format_lint(detail))

    assert(len(detail) == 2)
    assert (detail[0]['dsale'] == 'DSALE0001')
    assert (detail[1]['dsale'] == 'DSALE0002')
    # Check all the results

if __name__ == "__main__":
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main("-s -vv %s" % CURRENT_FILE)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import time
import datetime
import logging
import re
from lib389.utils import cmp

log = logging.getLogger(__name__)


class CSN(object):
    """CSN is Change Sequence Number
        csn.ts is the timestamp (time_t - seconds)
        csn.seq is the sequence number (max 65535)
        csn.rid is the replica ID of the originating supplier
        csn.subseq is not currently used"""

    csnpat = r'(.{8})(.{4})(.{4})(.{4})'
    csnre = re.compile(csnpat)

    def __init__(self, csnstr):
        match = CSN.csnre.match(csnstr)
        self.ts = 0
        self.seq = 0
        self.rid = 0
        self.subseq = 0
        if match:
            self.ts = int(match.group(1), 16)
            self.seq = int(match.group(2), 16)
            self.rid = int(match.group(3), 16)
            self.subseq = int(match.group(4), 16)
        elif csnstr:
            self.ts = 0
            self.seq = 0
            self.rid = 0
            self.subseq = 0
            log.info("%r is not a valid CSN" % csnstr)

    def csndiff(self, oth):
        return (oth.ts - self.ts,
                oth.seq - self.seq,
                oth.rid - self.rid,
                oth.subseq - self.subseq)

    def __cmp__(self, oth):
        if self is oth:
            return 0
        (tsdiff, seqdiff, riddiff, subseqdiff) = self.csndiff(oth)

        diff = tsdiff or seqdiff or riddiff or subseqdiff
        ret = 0
        if diff > 0:
            ret = 1
        elif diff < 0:
            ret = -1
        return ret

    def __eq__(self, oth):
        return cmp(self, oth) == 0

    def diff2str(self, oth):
        retstr = ''
        diff = oth.ts - self.ts
        if diff > 0:
            td = datetime.timedelta(seconds=diff)
            retstr = "is behind by %s" % td
        elif diff < 0:
            td = datetime.timedelta(seconds=-diff)
            retstr = "is ahead by %s" % td
        else:
            diff = oth.seq - self.seq
            if diff:
                retstr = "seq differs by %d" % diff
            elif self.rid != oth.rid:
                retstr = "rid %d not equal to rid %d" % (self.rid, oth.rid)
            else:
                retstr = "equal"
        return retstr

    def get_time_lag(self, oth):
        diff = oth.ts - self.ts
        if diff < 0:
            lag = datetime.timedelta(seconds=-diff)
        else:
            lag = datetime.timedelta(seconds=diff)
        return "{:0>8}".format(str(lag))

    def __repr__(self):
        return ("%s seq: %s rid: %s" % (time.strftime("%x %X", time.localtime(self.ts)),
                                        str(self.seq), str(self.rid)))

    def __str__(self):
        return self.__repr__()


class RUV(object):
    """RUV is Replica Update Vector
        ruv.gen is the generation CSN
        ruv.rid[1] through ruv.rid[N] are dicts - the number (1-N)
        is the replica ID
          ruv.rid[N][url] is the purl
          ruv.rid[N][min] is the min csn
          ruv.rid[N][max] is the max csn
          ruv.rid[N][lastmod] is the last modified timestamp
        example ruv attr:
        nsds50ruv: {replicageneration} 3b0ebc7f000000010000
        nsds50ruv: {replica 1 ldap://myhost:51010}
                    3b0ebc9f000000010000 3b0ebef7000000010000
        nsre_ruvplicaLastModified: {replica 1 ldap://myhost:51010}
                                               292398402093
        if the try repl flag is true, if getting the ruv from the
        suffix fails, try getting the ruv from the cn=replica entry
    """

    pre_gen = r'\{replicageneration\}\s+(\w+)'
    re_gen = re.compile(pre_gen)
    pre_ruv = r'\{replica\s+(\d+)\s+(.+?)\}\s*(\w*)\s*(\w*)'
    re_ruv = re.compile(pre_ruv)

    def __init__(self, ent):
        # rid is a dict
        # key is replica ID - val is dict of url, min csn, max csn
        self.rid = {}
        for item in ent.getValues('nsds50ruv'):
            matchgen = RUV.re_gen.match(item)
            matchruv = RUV.re_ruv.match(item)
            if matchgen:
                self.gen = CSN(matchgen.group(1))
            elif matchruv:
                rid = int(matchruv.group(1))
                self.rid[rid] = {'url': matchruv.group(2),
                                 'min': CSN(matchruv.group(3)),
                                 'max': CSN(matchruv.group(4))}
            else:
                log.info("unknown RUV element %r" % item)
        for item in ent.getValues('nsre_ruvplicaLastModified'):
            matchruv = RUV.re_ruv.match(item)
            if matchruv:
                rid = int(matchruv.group(1))
                self.rid[rid]['lastmod'] = int(matchruv.group(3), 16)
            else:
                log.info("unknown nsre_ruvplicaLastModified item %r" % item)

    def __cmp__(self, oth):
        if self is oth:
            return 0
        if not self:
            return -1  # None is less than something
        if not oth:
            return 1  # something is greater than None
        diff = cmp(self.gen, oth.gen)
        if diff:
            return diff
        for rid in list(self.rid.keys()):
            for item in ('max', 'min'):
                csn = self.rid[rid][item]
                csnoth = oth.rid[rid][item]
                diff = cmp(csn, csnoth)
                if diff:
                    return diff
        return 0

    def __eq__(self, oth):
        return cmp(self, oth) == 0

    def __str__(self):
        ret = 'generation: %s\n' % self.gen
        for rid in list(self.rid.keys()):
            ret = ret + 'rid: %s url: %s min: [%s] max: [%s]\n' % \
                (rid, self.rid[rid]['url'], self.rid[rid].get('min', ''),
                 self. rid[rid].get('max', ''))
        return ret

    def getdiffs(self, oth):
        """Compare two ruvs and return the differences
        returns a tuple - the first element is the
        result of cmp() - the second element is a string"""
        if self is oth:
            return (0, "\tRUVs are the same")
        if not self:
            return (-1, "\tfirst RUV is empty")
        if not oth:
            return (1, "\tsecond RUV is empty")
        diff = cmp(self.gen, oth.gen)
        if diff:
            return (diff, "\tgeneration [" + str(self.gen) +
                    "] not equal to [" + str(oth.gen) +
                    "]: likely not yet initialized")
        retstr = ''
        for rid in list(self.rid.keys()):
            for item in ('max', 'min'):
                csn = self.rid[rid][item]
                csnoth = oth.rid[rid][item]
                csndiff = cmp(csn, csnoth)
                if csndiff:
                    if len(retstr):
                        retstr += "\n"
                    retstr += ("\trid %d %scsn %s\n\t[%s] vs [%s]" %
                               (rid, item, csn.diff2str(csnoth), csn, csnoth))
                    if not diff:
                        diff = csndiff
        if not diff:
            retstr = "\tup-to-date - RUVs are equal"
        return (diff, retstr)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2023 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389.plugins import USNPlugin
from lib389.cli_conf import add_generic_plugin_parsers
from lib389.cli_base import CustomHelpFormatter


def display_usn_mode(inst, basedn, log, args):
    plugin = USNPlugin(inst)
    if plugin.is_global_mode_set():
        log.info("USN global mode is enabled")
    else:
        log.info("USN global mode is disabled")


def enable_global_mode(inst, basedn, log, args):
    plugin = USNPlugin(inst)
    plugin.enable_global_mode()
    log.info("USN global mode enabled")


def disable_global_mode(inst, basedn, log, args):
    plugin = USNPlugin(inst)
    plugin.disable_global_mode()
    log.info("USN global mode disabled")


def tombstone_cleanup(inst, basedn, log, args):
    plugin = USNPlugin(inst)
    log.info('Attempting to add task entry...')
    if not plugin.status():
        log.error("'%s' is disabled. Fix up task can't be executed" % plugin.rdn)
    task = plugin.cleanup(args.suffix, args.backend, args.max_usn)
    task.wait(timeout=args.timeout)
    exitcode = task.get_exit_code()
    if exitcode != 0:
        if exitcode is None:
            raise ValueError('USM tombstone cleanup task has not completed. Please, check logs')
        else:
            raise ValueError('USM tombstone cleanup task has failed. Please, check logs')
    else:
        log.info('Successfully added task entry')


def create_parser(subparsers):
    usn_parser = subparsers.add_parser('usn', help='Manage and configure USN plugin', formatter_class=CustomHelpFormatter)
    subcommands = usn_parser.add_subparsers(help='action')
    add_generic_plugin_parsers(subcommands, USNPlugin)

    global_mode_parser = subcommands.add_parser('global', help='Get or manage global USN mode (nsslapd-entryusn-global)', formatter_class=CustomHelpFormatter)
    global_mode_parser.set_defaults(func=display_usn_mode)
    global_mode_subcommands = global_mode_parser.add_subparsers(help='action')
    on_global_mode_parser = global_mode_subcommands.add_parser('on', help='Enables USN global mode', formatter_class=CustomHelpFormatter)
    on_global_mode_parser.set_defaults(func=enable_global_mode)
    off_global_mode_parser = global_mode_subcommands.add_parser('off', help='Disables USN global mode', formatter_class=CustomHelpFormatter)
    off_global_mode_parser.set_defaults(func=disable_global_mode)

    cleanup_parser = subcommands.add_parser('cleanup', help='Runs the USN tombstone cleanup task', formatter_class=CustomHelpFormatter)
    cleanup_parser.set_defaults(func=tombstone_cleanup)
    cleanup_group = cleanup_parser.add_mutually_exclusive_group(required=True)
    cleanup_group.add_argument('-s', '--suffix',
                               help='Sets the suffix or subtree in Directory Server to run the cleanup operation '
                                    'against. If the suffix is not specified, then the back end must be specified (suffix).')
    cleanup_group.add_argument('-n', '--backend',
                               help='Sets the Directory Server instance back end, or database, to run the cleanup '
                                    'operation against. If the back end is not specified, then the suffix must be '
                                    'specified. Backend instance in which USN tombstone entries (backend)')
    cleanup_parser.add_argument('-m', '--max-usn', type=int, help='Sets the highest USN value to delete when '
                                                                 'removing tombstone entries (max_usn_to_delete)')
    cleanup_parser.add_argument('--timeout', type=int, default=120,
                                help="Sets the cleanup task timeout.  Default is 120 seconds,")



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
"""Helpers that read instance configuration from ``dse.ldif`` (or related paths)."""

from lib389.dseldif import DSEldif


def get_ldapurl_from_serverid(instance):
    """Take an instance name, and get the host/port/protocol from dse.ldif
    and return a LDAP URL to use in the CLI tools (dsconf)

    :param instance: The server ID of a server instance
    :return tuple of LDAPURL and certificate directory (for LDAPS)
    """
    try:
        dse_ldif = DSEldif(None, instance)
    except Exception:
        return (None, None)

    port = dse_ldif.get("cn=config", "nsslapd-port", single=True)
    secureport = dse_ldif.get("cn=config", "nsslapd-secureport", single=True)
    host = dse_ldif.get("cn=config", "nsslapd-localhost", single=True)
    sec = dse_ldif.get("cn=config", "nsslapd-security", single=True)
    ldapi_listen = dse_ldif.get("cn=config", "nsslapd-ldapilisten", single=True)
    ldapi_autobind = dse_ldif.get("cn=config", "nsslapd-ldapiautobind", single=True)
    ldapi_socket = dse_ldif.get("cn=config", "nsslapd-ldapifilepath", single=True)
    certdir = dse_ldif.get("cn=config", "nsslapd-certdir", single=True)

    if ldapi_listen is not None and ldapi_listen.lower() == "on" and \
       ldapi_autobind is not None and ldapi_autobind.lower() == "on" and \
       ldapi_socket is not None:
        # Use LDAPI
        socket = ldapi_socket.replace("/", "%2f")  # Escape the path
        return ("ldapi://" + socket, None)
    elif sec is not None and sec.lower() == "on" and secureport is not None:
        # Use LDAPS
        return ("ldaps://{}:{}".format(host, secureport), certdir)
    else:
        # Use LDAP
        return ("ldap://{}:{}".format(host, port), None)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
from lib389._constants import *
from lib389 import DirSrv, Entry
import pytest
from lib389.utils import ensure_bytes, ensure_str

from lib389.mappingTree import MappingTrees, MappingTree

import sys
MAJOR, MINOR, _, _, _ = sys.version_info

INSTANCE_PORT = 54321
INSTANCE_SERVERID = 'standalone'

DEBUGGING = True

class TopologyStandalone(object):
    def __init__(self, standalone):
        standalone.open()
        self.standalone = standalone


@pytest.fixture(scope="module")
def topology(request):
    standalone = DirSrv(verbose=DEBUGGING)
    standalone.log.debug("Instance allocated")
    args = {SER_HOST: LOCALHOST,
            SER_PORT: INSTANCE_PORT,
            # SER_DEPLOYED_DIR:  INSTANCE_PREFIX,
            SER_SERVERID_PROP: INSTANCE_SERVERID}
    standalone.allocate(args)
    if standalone.exists():
        standalone.delete()
    standalone.create()
    standalone.open()

    def fin():
        if not DEBUGGING:
            standalone.delete()
    request.addfinalizer(fin)

    return TopologyStandalone(standalone)

def test_mappingtree(topology):

    mts = MappingTrees(topology.standalone)
    mt = mts.create(properties={
        'cn': ["dc=newexample,dc=com",],
        'nsslapd-state' : 'backend',
        'nsslapd-backend' : 'someRoot',
        })

    rmt = mts.get('someRoot')
    rmt.delete()






# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2018 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import os
import io
import sys
import pytest
import time
from lib389.cli_conf.chaining import (config_get, config_set, def_config_get, def_config_set,
                                      create_link, get_link, edit_link, delete_link,
                                      monitor_link, list_links)
from lib389.cli_conf.backend import (backend_create, backend_delete, backend_list)
from lib389.cli_base import LogCapture, FakeArgs
from lib389.tests.cli import check_output
from lib389.tests.topologies import topology_st

from lib389.utils import ds_is_older
pytestmark = pytest.mark.skipif(ds_is_older('1.4.0'), reason="Not implemented")

LINK_NAME = "mylink"
LINK_SUFFIX = "ou=link,dc=chaining,dc=test"
SUFFIX = "dc=chaining,dc=test"
BE_NAME = "chainingRoot"
stdout = sys.__stdout__


@pytest.fixture(scope="function")
def create_backend(topology_st, request):
    """Create backend "dc=backend,dc=test" / backendRoot
    """
    sys.stdout = io.StringIO()

    args = FakeArgs()
    args.cn = BE_NAME
    args.be_name = BE_NAME
    args.suffix = False
    args.nsslapd_suffix = SUFFIX
    args.skip_subsuffixes = False
    args.json = False
    args.parent_suffix = False
    args.create_entries = True
    args.suffix = SUFFIX
    backend_create(topology_st.standalone, None, None, args)
    check_output("The database was successfully created")

    def fin():
        sys.stdout = io.StringIO()

        args = FakeArgs()
        args.cn = BE_NAME
        args.be_name = BE_NAME
        args.suffix = SUFFIX
        args.skip_subsuffixes = False
        args.json = False

        # Delete backend
        backend_delete(topology_st.standalone, None, None, args, warn=False)
        check_output("successfully deleted")

        # Verify it's removed
        args.suffix = False
        backend_list(topology_st.standalone, None, None, args)
        check_output(BE_NAME, missing=True)

    request.addfinalizer(fin)


def test_chaining_cli(topology_st, create_backend):
    """Test creating, listing, getting, and deleting a backend (and subsuffix)

    :id: 800f432a-52ab-4661-ac66-a2bdd9b984d7
    :setup: Standalone instance
    :steps:
        1. Update config controls and components
        2. Verify update to config
        3. Set default config
        4. Verify update to default config
        5. Add DB Link
        6. Verify Link was created
        7. Edit Link
        8. Verify edit to link
        9. Test monitor
        10. Delete link
        11. Verify link was deleted
    :expectedresults:
        1. Success
        2. Success
        3. Success
        4. Success
        5. Success
        6. Success
        7. Success
        8. Success
        9. Success
        10. Success
        11. Success
    """
    topology_st.logcap = LogCapture()
    sys.stdout = io.StringIO()
    args = FakeArgs()
    args.CHAIN_NAME = [LINK_NAME]
    args.suffix = LINK_SUFFIX
    args.json = False
    args.add_control = None
    args.del_control = None
    args.add_comp = None
    args.del_comp = None

    # Set config (add control)
    args.add_control = '1.1.1.1.1.1.1'
    config_set(topology_st.standalone, None, None, args)
    args.add_control = None
    check_output("updated chaining configuration")

    # Verify config change
    config_get(topology_st.standalone, None, None, args)
    check_output("1.1.1.1.1.1.1")

    # Set config (delete control)
    args.del_control = '1.1.1.1.1.1.1'
    config_set(topology_st.standalone, None, None, args)
    args.del_control = None
    check_output("updated chaining configuration")

    # Verify config change
    config_get(topology_st.standalone, None, None, args)
    check_output("1.1.1.1.1.1.1", missing=True)

    # Set config (add comp)
    args.add_comp = 'cn=test,cn=config'
    config_set(topology_st.standalone, None, None, args)
    args.add_comp = None
    check_output("updated chaining configuration")

    # Verify config change
    config_get(topology_st.standalone, None, None, args)
    check_output('cn=test,cn=config')

    # Set config (delete comp)
    args.del_comp = 'cn=test,cn=config'
    config_set(topology_st.standalone, None, None, args)
    args.del_comp = None
    check_output("updated chaining configuration")

    # Verify config change
    config_get(topology_st.standalone, None, None, args)
    check_output("cn=test,cn=config", missing=True)

    # Set default config
    args.time_limit = '5555'
    def_config_set(topology_st.standalone, None, None, args)
    check_output("updated chaining default instance creation configuration")

    # Verify default config change
    def_config_get(topology_st.standalone, None, None, args)
    check_output("nsslapd_timelimit: 5555")

    # Create database link
    args.server_url = "ldap://localhost.localdomain"
    args.bind_dn = "cn=link_admin," + SUFFIX
    args.bind_pw = "secret_157"
    args.bind_mech = "LDAP"
    create_link(topology_st.standalone, None, None, args)
    check_output("created database link")

    # Verify link was created
    list_links(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output(LINK_NAME)

    # Edit link
    args.bind_dn = "uid=newuser,cn=config"
    args.suffix = None
    edit_link(topology_st.standalone, None, None, args)
    check_output("updated database chaining link")

    # Verify link was edited
    args.cn = LINK_NAME
    get_link(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output("uid=newuser,cn=config")

    # Test monitor
    time.sleep(2)  # need time for link to start up and generate monitor
    monitor_link(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output("nssearchonelevelcount: ")

    # Delete link
    delete_link(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output("deleted database link")

    # Verify link was deleted
    list_links(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output(LINK_NAME, missing=True)

    # Done!



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import pytest
import logging
import ldap
import sys

from lib389.cli_base.dsrc import dsrc_to_ldap

MAJOR, MINOR, _, _, _ = sys.version_info
if MAJOR >= 3:
    import configparser

log = logging.getLogger(__name__)

def write_inf(content):
    with open('/tmp/dsrc_test.inf', 'w') as f:
        f.write(content)

@pytest.mark.skipif(sys.version_info < (3,0), reason="requires python3")
def test_dsrc_single_section():
    # Write out inf.
    write_inf("""
[localhost]
    """)
    # Parse it and assert the content
    with pytest.raises(configparser.NoOptionError):
        i = dsrc_to_ldap('/tmp/dsrc_test.inf', 'localhost', log)

    write_inf("""
[localhost]
uri = ldaps://localhost:636
    """)
    # Parse it and assert the content
    i = dsrc_to_ldap('/tmp/dsrc_test.inf', 'localhost', log)
    assert(i == {
            'uri': 'ldaps://localhost:636',
            'basedn': None,
            'binddn': None,
            'saslmech': None,
            'tls_cacertdir': None,
            'tls_cert': None,
            'tls_key': None,
            'tls_reqcert': ldap.OPT_X_TLS_HARD,
            'starttls': False}
        )

    write_inf("""
[localhost]
uri = ldaps://localhost:636
basedn = dc=example,dc=com
binddn = cn=Directory Manager
starttls = true
    """)
    # Parse it and assert the content
    i = dsrc_to_ldap('/tmp/dsrc_test.inf', 'localhost', log)
    assert(i == {
            'uri': 'ldaps://localhost:636',
            'basedn': 'dc=example,dc=com',
            'binddn': 'cn=Directory Manager',
            'saslmech': None,
            'tls_cacertdir': None,
            'tls_cert': None,
            'tls_key': None,
            'tls_reqcert': ldap.OPT_X_TLS_HARD,
            'starttls': True}
        )

@pytest.mark.skipif(sys.version_info < (3,0), reason="requires python3")
def test_dsrc_two_section():
    write_inf("""
[localhost]
uri = ldaps://localhost:636

[localhost2]
uri = ldaps://localhost:6362
    """)
    # Parse it and assert the content
    i = dsrc_to_ldap('/tmp/dsrc_test.inf', 'localhost', log)
    assert(i == {
            'uri': 'ldaps://localhost:636',
            'basedn': None,
            'binddn': None,
            'saslmech': None,
            'tls_cacertdir': None,
            'tls_cert': None,
            'tls_key': None,
            'tls_reqcert': ldap.OPT_X_TLS_HARD,
            'starttls': False}
        )

    i = dsrc_to_ldap('/tmp/dsrc_test.inf', 'localhost2', log)
    assert(i == {
            'uri': 'ldaps://localhost:6362',
            'basedn': None,
            'binddn': None,
            'saslmech': None,
            'tls_cacertdir': None,
            'tls_cert': None,
            'tls_key': None,
            'tls_reqcert': ldap.OPT_X_TLS_HARD,
            'starttls': False}
        )

    # Section doesn't exist
    i = dsrc_to_ldap('/tmp/dsrc_test.inf', 'localhost3', log)
    assert(i is None)

@pytest.mark.skipif(sys.version_info < (3,0), reason="requires python3")
def test_dsrc_reqcert():
    write_inf("""
[localhost]
uri = ldaps://localhost:636
tls_reqcert = invalid
    """)
    # Parse it and assert the content
    with pytest.raises(Exception):
        i = dsrc_to_ldap('/tmp/dsrc_test.inf', 'localhost', log)

    write_inf("""
[localhost]
uri = ldaps://localhost:636
tls_reqcert = hard
    """)
    # Parse it and assert the content
    i = dsrc_to_ldap('/tmp/dsrc_test.inf', 'localhost', log)
    assert(i['tls_reqcert'] == ldap.OPT_X_TLS_HARD)

    write_inf("""
[localhost]
uri = ldaps://localhost:636
tls_reqcert = allow
    """)
    # Parse it and assert the content
    i = dsrc_to_ldap('/tmp/dsrc_test.inf', 'localhost', log)
    assert(i['tls_reqcert'] == ldap.OPT_X_TLS_ALLOW)

    write_inf("""
[localhost]
uri = ldaps://localhost:636
tls_reqcert = never
    """)
    # Parse it and assert the content
    i = dsrc_to_ldap('/tmp/dsrc_test.inf', 'localhost', log)
    assert(i['tls_reqcert'] == ldap.OPT_X_TLS_NEVER)

@pytest.mark.skipif(sys.version_info < (3,0), reason="requires python3")
def test_dsrc_saslmech():
    write_inf("""
[localhost]
uri = ldaps://localhost:636
saslmech = INVALID_MECH
    """)
    # Parse it and assert the content
    with pytest.raises(Exception):
        i = dsrc_to_ldap('/tmp/dsrc_test.inf', 'localhost', log)

    write_inf("""
[localhost]
uri = ldaps://localhost:636
saslmech = EXTERNAL
    """)
    # Parse it and assert the content
    i = dsrc_to_ldap('/tmp/dsrc_test.inf', 'localhost', log)
    assert(i['saslmech'] == 'EXTERNAL')

    write_inf("""
[localhost]
uri = ldaps://localhost:636
saslmech = PLAIN
    """)
    # Parse it and assert the content
    i = dsrc_to_ldap('/tmp/dsrc_test.inf', 'localhost', log)
    assert(i['saslmech'] == 'PLAIN')



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#

import pytest
import ldap

from lib389.tests.topologies import topology_st

from lib389.backend import Backends
from lib389.index import Indexes

def test_default_index_list(topology_st):
    indexes = Indexes(topology_st.standalone)
    # create and delete a default index.

    index = indexes.create(properties={
        'cn': 'modifytimestamp',
        'nsSystemIndex': 'false',
        'nsIndexType': 'eq'
        })
    default_index_list = indexes.list()
    found = False
    for i in default_index_list:
        if i.dn.startswith('cn=modifytimestamp'):
            found = True
    assert found
    index.delete()

    default_index_list = indexes.list()
    found = False
    for i in default_index_list:
        if i.dn.startswith('cn=modifytimestamp'):
            found = True
    assert not found

def test_backend_index(topology_st):
    backends = Backends(topology_st.standalone)
    ur_backend = backends.get('userRoot')
    ur_indexes = ur_backend.get_indexes()

    index = ur_indexes.create(properties={
        'cn': 'modifytimestamp',
        'nsSystemIndex': 'false',
        'nsIndexType': 'eq'
        })

    index_list = ur_indexes.list()
    found = False
    for i in index_list:
        if i.dn.startswith('cn=modifytimestamp'):
            found = True
    assert found
    index.delete()

    index_list = ur_indexes.list()
    found = False
    for i in index_list:
        if i.dn.startswith('cn=modifytimestamp'):
            found = True
    assert not found






# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2022 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap
from lib389._constants import DN_CHAIN
from lib389.properties import BACKEND_OBJECTCLASS_VALUE
from lib389.monitor import MonitorChaining
from lib389.mappingTree import MappingTrees
from lib389._mapped_object import DSLdapObjects, DSLdapObject
from lib389.backend import Backends
from lib389.rootdse import RootDSE


class ChainingConfig(DSLdapObject):
    """Chaining Config DSLdapObject with:
    - must attributes = ['cn']
    - RDN attribute is 'cn'

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    _must_attributes = ['cn']

    def __init__(self, instance, dn=None):
        super(ChainingConfig, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['cn']
        self._create_objectclasses = ['top', 'extensibleObject']
        self._protected = True
        self._dn = "cn=config,cn=chaining database,cn=plugins,cn=config"

    def get_controls(self):
        """Get a list of the supported controls from the root DSE entry
        :return list of OIDs
        """
        rootdse = RootDSE(self._instance)
        ctrls = rootdse.get_supported_ctrls()
        ctrls.sort()
        return ctrls

    def get_comps(self):
        """Return a list of the available plugin components
        :return list of plugin components
        """
        comps = self.get_attr_vals_utf8_l('nspossiblechainingcomponents')
        comps.sort()
        return comps


class ChainingDefault(DSLdapObject):
    """Chaining Default Config settings DSLdapObject with:
    - must attributes = ['cn']
    - RDN attribute is 'cn'

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    _must_attributes = ['cn']

    def __init__(self, instance, dn=None):
        super(ChainingDefault, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['cn']
        self._create_objectclasses = ['top', 'extensibleObject']
        self._protected = True
        self._dn = "cn=default instance config,cn=chaining database,cn=plugins,cn=config"


class ChainingLink(DSLdapObject):
    """Chaining Backend DSLdapObject with:
    - must attributes = ['cn', 'nsslapd-suffix', 'nsmultiplexorbinddn',
                         'nsmultiplexorcredentials', 'nsfarmserverurl'
    - RDN attribute is 'cn'

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    _must_attributes = ['nsslapd-suffix', 'cn']

    def __init__(self, instance, dn=None, rdn=None):
        super(ChainingLink, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['nsslapd-suffix', 'cn', 'nsmultiplexorbinddn',
                                 'nsmultiplexorcredentials', 'nsfarmserverurl']
        self._create_objectclasses = ['top', 'extensibleObject', BACKEND_OBJECTCLASS_VALUE]
        self._protected = False
        self._basedn = "cn=chaining database,cn=plugins,cn=config"
        self._mts = MappingTrees(self._instance)

    def get_monitor(self):
        """Get a MonitorChaining(DSLdapObject) for the chaining link
        :returns - chaining monitor entry
        """
        return MonitorChaining(instance=self._instance, dn="cn=monitor,%s" % self._dn)

    def del_link(self):
        """
        Remove the link from the parent suffix backend entry
        Delete chaining monitor entry
        Delete chaining entry
        """

        rdn = self.get_attr_val_utf8_l('nsslapd-suffix')
        try:
            mt = self._mts.get(selector=rdn)
            mt.delete()
        except ldap.NO_SUCH_OBJECT:
            # Righto, it's already gone! Do nothing ...
            pass

        # Delete the monitoring entry
        monitor = self.get_monitor()
        monitor.delete()

        # Delete the link
        self.delete()

    def create(self, rdn=None, properties=None, basedn=None):
        """Create the link entry, and the mapping tree entry(if needed)
        """

        # Create chaining entry
        link = super(ChainingLink, self).create(rdn, properties, basedn)

        # Create mapping tree entry
        dn_comps = ldap.explode_dn(properties['nsslapd-suffix'][0])
        parent_suffix = ','.join(dn_comps[1:])
        mt_properties = {
            'cn': properties['nsslapd-suffix'][0],
            'nsslapd-state': 'backend',
            'nsslapd-backend': properties['cn'][0],
            'nsslapd-parent-suffix': parent_suffix
        }
        try:
            self._mts.ensure_state(properties=mt_properties)
        except ldap.ALREADY_EXISTS:
            pass
        return link


class ChainingLinks(DSLdapObjects):
    """DSLdapObjects that represents DN_LDBM base DN
    This only does ldbm backends. Chaining backends are a special case
    of this, so they can be subclassed off.

    :param instance: An instance
    :type instance: lib389.DirSrv
    """

    def __init__(self, instance, basedn=None):
        # Basedn has to be here, despite not being used to satisfy
        # cli_base _generic_create.
        super(ChainingLinks, self).__init__(instance=instance)
        self._objectclasses = [BACKEND_OBJECTCLASS_VALUE]
        self._filterattrs = ['cn', 'nsslapd-suffix', 'nsslapd-directory']
        self._childobject = ChainingLink
        self._basedn = DN_CHAIN

    def add_link(self, props):
        """
        Create chaining entry
        Add link to parent suffix backend entry
        """
        self.create(properties=props)
        suffix = props['nsslapd-suffix'][0].lower()
        rdn = props['cn'][0]
        be_insts = Backends(self._instance).list()
        for be in be_insts:
            be_suffix = be.get_attr_val_utf8_l('nsslapd-suffix')
            if suffix == be_suffix:
                # Add chaining link as backend under the suffix
                be.add('nsslapd-backend', rdn)
